# -*- coding: utf-8 -*-

import sys

try:
    import simplejson as json
except (ImportError, SyntaxError):
    # simplejson does not support Python 3.2, it thows a SyntaxError
    # because of u'...' Unicode literals.
    import json  # noqa

try:
    from base64 import encodebytes
except ImportError:
    from base64 import encodestring as encodebytes

# -------
# Pythons
# -------

_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)


# ---------
# Specifics
# ---------

if is_py2:
    from urlparse import urlparse  # noqa
    import StringIO
    StringIO = BytesIO = StringIO.StringIO

elif is_py3:
    from urllib.parse import urlparse  # noqa
    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO



# coding:utf-8
from ks3 import utils

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2

try:
    from hashlib import sha1 as sha
    from hashlib import sha256 as sha256
except ImportError:
    import sha
    sha256 = None

import hmac
import datetime

def url_encode(key):
    if not key:
        return ""
    encode_key = parse.quote(utils.get_utf8_value(key))
    # if '%20' in encode_key:
    #     encode_key = encode_key.replace('%20', '+')
    #
    # if '%2A' in encode_key:
    #     encode_key = encode_key.replace('%2A', '*')

    if '%7E' in encode_key:
        encode_key = encode_key.replace('%7E', '~')

    if '%2F' in encode_key:
        encode_key = encode_key.replace('%2F', '/')

    return encode_key


def encode_params(query_args):
    if not query_args:
        return ""
    map_args = {}
    if isinstance(query_args, dict):
        map_args = query_args
    else:
        for param in query_args.split("&"):
            kv = param.split("=", 1)
            k = kv[0]
            if len(kv) == 1:
                map_args[k] = ""
            else:
                v = kv[1]
                map_args[k] = v
    if not map_args:
        return ""
    sorted_keys = list(map_args.keys())
    sorted_keys.sort()
    buf_list = []
    for k in sorted_keys:
        v = map_args.get(k)
        if v:
            buf_list.append("%s=%s" % (k, v))
        else:
            # 与v2签名不同，v4签名中，参数值为空时，也需要加上=号，以通过kop鉴权
            buf_list.append("%s=" % k)
    return "&".join(buf_list)

# Key derivation functions. See:
# http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
def sign(key, msg):
    return hmac.new(key, msg.encode('utf-8'), sha256).digest()

def getSignatureKey(key, dateStamp, regionName, serviceName):
    kDate = sign(('AWS4' + key).encode('utf-8'), dateStamp)
    kRegion = sign(kDate, regionName)
    kService = sign(kRegion, serviceName)
    kSigning = sign(kService, 'aws4_request')
    return kSigning

def getCanonicalRequest(host, method, signed_headers, amzdate, request_parameters, body=""):
    # 正规化 uri；use '/' if no path
    canonical_uri = '/'
    # 正规化 headers
    canonical_headers = 'host:' + host + '\n' + 'x-amz-date:' + amzdate + '\n'
    # must be URL-encoded (space=%20)
    canonical_querystring = request_parameters
    # Create payload hash (hash of the request body content). For GET
    # requests, the payload is an empty string ("").
    payload_hash = sha256(body.encode('utf-8')).hexdigest()

    canonical_request = method + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + canonical_headers + '\n' + signed_headers + '\n' + payload_hash
    return canonical_request

def add_auth_header(access_key_id, secret_access_key, region, service, host, method, query_args, body, headers=None, tm=datetime.datetime.utcnow()):
    # "Host" and "x-amz-date" are always required.
    if headers is None:
        headers = {}
    signed_headers = 'host;x-amz-date'
    if not access_key_id:
        return
    amzdate = tm.strftime('%Y%m%dT%H%M%SZ')
    if 'x-amz-date' not in headers:
        headers['x-amz-date'] = amzdate

    datestamp = tm.strftime('%Y%m%d')
    signing_key = getSignatureKey(secret_access_key, datestamp, region, service)
    # print('signing_key: ', signing_key)
    canonical_request = getCanonicalRequest(host, method, signed_headers, amzdate, encode_params(query_args), body)

    # print('canonical_request', canonical_request)
    algorithm = 'AWS4-HMAC-SHA256'
    credential_scope = datestamp + '/' + region + '/' + service + '/' + 'aws4_request'
    string_to_sign = algorithm + '\n' + amzdate + '\n' + credential_scope + '\n' + sha256(
        canonical_request.encode('utf-8')).hexdigest()
    # print('string_to_sign', string_to_sign)
    signature = hmac.new(signing_key, (string_to_sign).encode('utf-8'), sha256).hexdigest()
    # print('signature: ', signature)
    headers['Authorization'] = algorithm + ' ' + 'Credential=' + access_key_id + '/' + credential_scope + ', ' + 'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature
    # print('authorization: ', algorithm + ' ' + 'Credential=' + access_key_id + '/' + credential_scope + ', ' + 'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature)
    return headers


# -*- coding: utf-8 -*-
import base64
import email
import hashlib
import logging
import math
import mimetypes
import os
import re
import tempfile
import xml
from hashlib import md5

import requests

from ks3 import handler, utils
from ks3.auth import url_encode
from ks3.copy import CopyTask, AcrossRegionCopyTask
from ks3.download import DownloadTask
from ks3.objectTagging import ObjectTagging
from ks3.responseResult import ResponseMetadata, ResponseResult
from ks3.upload import UploadTask
from ks3.xmlParsers.objectMigration import MigrationResult, MigrationConfiguration

try:
    import urllib.parse as parse
except ImportError:
    import urllib as parse
import binascii

import errno
from ks3.exception import KS3ClientError, S3ResponseError, ParamValidationError
from ks3.exception import StorageDataError, PleaseRetryException
from ks3.keyfile import KeyFile
from ks3.user import User
from ks3.compat import BytesIO
from ks3.utils import compute_md5, compute_encrypted_md5, BlockThreadPoolExecutor, DEFAULT_PART_SIZE, GB, MAX_SIZE
from ks3.utils import find_matching_headers, merge_meta, merge_headers_by_name

try:
    from ks3.encryption import Crypts
    from ks3.encryptFp import EncryptFp
except:
    pass
import aiofiles


logger = logging.getLogger(__name__)


class Key(object):
    DefaultContentType = 'application/octet-stream'

    RestoreBody = """<?xml version="1.0" encoding="UTF-8"?>
      <RestoreRequest xmlns="http://s3.ksyun.com">
        <Days>%s</Days>
      </RestoreRequest>"""

    # BufferSize = ks3.config.getint('Boto', 'key_buffer_size', 8192)
    BufferSize = 8192

    # The object metadata fields a user can set, other than custom metadata
    # fields (i.e., those beginning with a provider-specific prefix like
    # x-amz-meta).
    base_user_settable_fields = set(["cache-control", "content-disposition",
                                     "content-encoding", "content-language",
                                     "content-md5", "content-type",
                                     "x-robots-tag", "expires"])
    _underscore_base_user_settable_fields = set()
    for f in base_user_settable_fields:
        _underscore_base_user_settable_fields.add(f.replace('-', '_'))
    # Metadata fields, whether user-settable or not, other than custom
    # metadata fields (i.e., those beginning with a provider specific prefix
    # like x-amz-meta).
    base_fields = (base_user_settable_fields |
                   set(["last-modified", "content-length", "date", "etag"]))

    def __init__(self, bucket=None, name=None, *args, **kwargs):
        self.bucket = bucket
        self.name = name
        self.metadata = {}
        self.cache_control = None
        self.content_type = self.DefaultContentType
        self.content_encoding = None
        self.content_disposition = None
        self.content_language = None
        self.filename = None
        self.etag = None
        self.is_latest = False
        self.last_modified = None
        self.owner = None
        self.storage_class = None
        self.path = None
        self.resp = None
        self.mode = None
        self._size = None
        self.version_id = None
        self.source_version_id = None
        self.delete_marker = False
        self.encrypted = None
        self.retention_id = None
        self.recycle_time = None
        self.estimated_clear_time = None
        #
        self.local_hashes = {}
        self.user_meta = {}
        self.tagging_count = None
        self.object_type = None
        self.append_next_position = None
        self.server_crc = None
        self.client_crc = '0'

        self.response_metadata = ResponseMetadata(**kwargs)

    @property
    def size(self):
        if self._size is None and self.bucket:
            key = self.bucket.get_key(self.name, validate=True)
            if key:
                self._size = key.size
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    def __repr__(self):
        if self.bucket:
            name = '<Key: %s,%s>' % (self.bucket.name, self.name)
        else:
            name = '<Key: None,%s>' % self.name

        # Encode to bytes for Python 2 to prevent display decoding issues
        if not isinstance(name, str):
            name = name.encode('utf-8')

        return name

    def __iter__(self):
        return self

    @property
    def provider(self):
        provider = None
        if self.bucket and self.bucket.connection:
            provider = self.bucket.connection.provider
        return provider

    def _get_key(self):
        return self.name

    def _set_key(self, value):
        self.name = value

    def next(self):
        """
        By providing a next method, the key object supports use as an iterator.
        For example, you can now say:

        for bytes in key:
            write bytes to a file or whatever

        All of the HTTP connection stuff is handled for you.
        """
        self.open_read()
        data = self.resp.read(self.BufferSize)
        if not data:
            self.close()
            raise StopIteration
        return data

    # Python 3 iterator support
    __next__ = next

    def read(self, size=0):
        self.open_read()
        if size == 0:
            data = self.resp.read()
        else:
            data = self.resp.read(size)

        if self.bucket.connection.enable_crc and data:
            crc_calculater = utils.Crc64(init_crc=int(self.client_crc))
            crc_calculater.update(data)
            self.client_crc = str(crc_calculater.crc)

        if not data:
            self.close()
        return data

    def startElement(self, name, attrs, connection):
        if name == 'Owner':
            self.owner = User(self)
            return self.owner
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Key':
            self.name = value
        elif name == 'ETag':
            self.etag = value
        elif name == 'IsLatest':
            if value == 'true':
                self.is_latest = True
            else:
                self.is_latest = False
        elif name == 'LastModified':
            self.last_modified = value
        elif name == 'Size':
            self.size = int(value)
        elif name == 'StorageClass':
            self.storage_class = value
        elif name == 'Owner':
            pass
        elif name == 'VersionId':
            self.version_id = value
        elif name == 'RetentionId':
            self.retention_id = value
        elif name == 'RecycleTime':
            self.recycle_time = value
        elif name == 'EstimatedClearTime':
            self.estimated_clear_time = value
        else:
            setattr(self, name, value)

    def delete(self, headers=None):
        """
        Delete this key from S3
        """
        return self.bucket.delete_key(self.name, version_id=self.version_id,
                                      headers=headers)

    def generate_url(self, expires_in, method='GET', headers=None,
                     query_auth=True, force_http=False, response_headers=None,
                     expires_in_absolute=False, version_id=None,
                     policy=None, reduced_redundancy=False, encrypt_key=False, domain=False, image_attrs=None):
        """
        Generate a URL to access this key.
        """
        provider = self.bucket.connection.provider
        version_id = version_id or self.version_id
        if headers is None:
            headers = {}
        else:
            headers = headers.copy()

        # add headers accordingly (usually PUT case)
        # if policy:
        #    headers[provider.acl_header] = policy
        # if reduced_redundancy:
        #    self.storage_class = 'REDUCED_REDUNDANCY'
        #    if provider.storage_class_header:
        #        headers[provider.storage_class_header] = self.storage_class
        # if encrypt_key:
        #    headers[provider.server_side_encryption_header] = 'AES256'
        # headers = merge_meta(headers, self.metadata, provider)

        name = self.name
        if image_attrs is not None:
            name = name + image_attrs

        url = self.bucket.connection.generate_url(expires_in, method,
                                                  self.bucket.name, name,
                                                  headers, query_auth,
                                                  force_http,
                                                  response_headers,
                                                  expires_in_absolute,
                                                  version_id)
        if domain is True:
            url = url.replace('http://' + self.bucket.name + '.', 'http://')
            url = url.replace('https://' + self.bucket.name + '.', 'https://')

        return url

    # GetObjectPresignedUrl 生成临时上传链接
    def get_presigned_url(self, expires_in, headers=None, expires_in_absolute=False, version_id=None, domain=False):
        """
        生成临时上传链接
        """
        version_id = version_id or self.version_id
        if headers is None:
            headers = {}
        else:
            headers = headers.copy()

        url = self.bucket.connection.generate_url(expires_in, 'PUT',
                                                  self.bucket.name, self.name,
                                                  headers=headers, expires_in_absolute=expires_in_absolute,
                                                  version_id=version_id)
        if domain is True:
            url = url.replace('http://' + self.bucket.name + '.', 'http://')
            url = url.replace('https://' + self.bucket.name + '.', 'https://')

        return url

    def should_retry(self, response, chunked_transfer=False):
        provider = self.bucket.connection.provider

        if not chunked_transfer:
            if response.status in [500, 503]:
                # 500 & 503 can be plain retries.
                return True

            if response.getheader('location'):
                # If there's a redirect, plain retry.
                return True

        if 200 <= response.status <= 299:
            self.etag = response.getheader('etag')
            md5 = self.md5
            if isinstance(md5, bytes):
                md5 = md5.decode('utf-8')

            # If you use customer-provided encryption keys, the ETag value that
            # Amazon S3 returns in the response will not be the MD5 of the
            # object.
            server_side_encryption_customer_algorithm = response.getheader(
                'x-amz-server-side-encryption-customer-algorithm', None)
            if server_side_encryption_customer_algorithm is None:
                if self.etag != '"%s"' % md5:
                    raise provider.storage_data_error(
                        'ETag from S3 did not match computed MD5. '
                        '%s vs. %s' % (self.etag, self.md5))

            return True

        if response.status == 400:
            # The 400 must be trapped so the retry handler can check to
            # see if it was a timeout.
            # If ``RequestTimeout`` is present, we'll retry. Otherwise, bomb
            # out.
            body = response.read()
            err = provider.storage_response_error(
                response.status,
                response.reason,
                body
            )

            if err.error_code in ['RequestTimeout']:
                raise PleaseRetryException(
                    "Saw %s, retrying" % err.error_code,
                    response=response
                )

        return False

    def send_file(self, fp, method="PUT", headers=None, cb=None, num_cb=10,
                  query_args=None, chunked_transfer=False, size=None,
                  crypt_context=None):
        """
        Upload a file to a key into a bucket on S3.
        """
        resp = self._send_file_internal(fp, method=method, headers=headers, cb=cb, num_cb=num_cb,
                                        query_args=query_args, chunked_transfer=chunked_transfer,
                                        size=size, crypt_context=crypt_context)
        resp.read()
        return resp

    def _send_file_internal(self, fp, method="PUT", headers=None, cb=None, num_cb=10,
                            query_args=None, chunked_transfer=False, size=None,
                            hash_algs=None, crypt_context=None):
        provider = self.bucket.connection.provider
        # try:
        #     spos = fp.tell()
        # except IOError:
        #     spos = None
        #     self.read_from_stream = False

        # If hash_algs is unset and the MD5 hasn't already been computed,
        # default to an MD5 hash_alg to hash the data on-the-fly.
        # if hash_algs is None and not self.md5:
        #     hash_algs = {'md5': md5}
        # digesters = dict((alg, hash_algs[alg]()) for alg in hash_algs or {})

        if not headers:
            headers = {}
        else:
            headers = headers.copy()
        # Overwrite user-supplied user-agent.
        for header in find_matching_headers('User-Agent', headers):
            del headers[header]
        # headers['User-Agent'] = UserAgent
        # If storage_class is None, then a user has not explicitly requested
        # a storage class, so we can assume STANDARD here
        if self.storage_class not in [None, 'STANDARD']:
            headers[provider.storage_class_header] = self.storage_class
        if find_matching_headers('Content-Encoding', headers):
            self.content_language = merge_headers_by_name(
                'Content-Language', headers)
        content_type_headers = find_matching_headers('Content-Type', headers)
        if content_type_headers:
            # Some use cases need to suppress sending of the Content-Type
            # header and depend on the receiving server to set the content
            # type. This can be achieved by setting headers['Content-Type']
            # to None when calling this method.
            if (len(content_type_headers) == 1 and
                    headers[content_type_headers[0]] is None):
                # Delete null Content-Type value to skip sending that header.
                del headers[content_type_headers[0]]
            else:
                self.content_type = merge_headers_by_name(
                    'Content-Type', headers)
        elif self.path:
            self.content_type = mimetypes.guess_type(self.path)[0]
            if self.content_type is None:
                self.content_type = self.DefaultContentType
            headers['Content-Type'] = self.content_type
        else:
            headers['Content-Type'] = self.content_type
        if self.base64md5:
            headers['Content-MD5'] = self.base64md5
        if chunked_transfer:
            headers['Transfer-Encoding'] = 'chunked'
            # if not self.base64md5:
            #    headers['Trailer'] = "Content-MD5"
        else:
            headers['Content-Length'] = str(self.size)
        # This is terrible. We need a SHA256 of the body for SigV4, but to do
        # the chunked ``sender`` behavior above, the ``fp`` isn't available to
        # the auth mechanism (because closures). Detect if it's SigV4 & embelish
        # while we can before the auth calculations occur.
        # if 'hmac-v4-s3' in self.bucket.connection._required_auth_capability():
        #    kwargs = {'fp': fp, 'hash_algorithm': hashlib.sha256}
        #    if size is not None:
        #        kwargs['size'] = size
        #    headers['_sha256'] = compute_hash(**kwargs)[0]
        # headers['Expect'] = '100-Continue'
        headers = merge_meta(headers, self.metadata, provider)
        if self.size < 0 or self.size > MAX_SIZE:
            raise ParamValidationError('Key size={0} should between 0 and {1} bytes'.format(self.size, MAX_SIZE))
        # keyFile类型没有本地加密
        if self.bucket.connection.local_encrypt and crypt_context and not isinstance(fp, KeyFile):
            if crypt_context.action_info == "put":
                fp = EncryptFp(fp, crypt_context, "put")
                md5_generator = hashlib.md5()
                md5_generator.update(crypt_context.key)
                headers["x-kss-meta-key"] = base64.b64encode(md5_generator.hexdigest().encode()).decode()
                headers["x-kss-meta-iv"] = base64.b64encode(crypt_context.first_iv).decode()
            if crypt_context.action_info == "upload_part":
                if crypt_context.part_num == 1:
                    fp = EncryptFp(fp, crypt_context, "upload_part", isUploadFirstPart=True,
                                   isUploadLastPart=crypt_context.is_last_part)
                else:
                    fp = EncryptFp(fp, crypt_context, "upload_part", isUploadFirstPart=False,
                                   isUploadLastPart=crypt_context.is_last_part)
            if self.base64md5:
                headers["x-kss-meta-unencrypted-content-md5"] = self.base64md5
            if headers.get("Content-MD5"):
                headers.pop("Content-MD5")
            if crypt_context.calc_md5:
                headers['Content-MD5'] = self.compute_encrypted_md5(fp)
                fp.block_count = 0
                fp.calc_iv = ""
            if chunked_transfer:
                headers['Transfer-Encoding'] = 'chunked'
            else:
                headers['Content-Length'] = str(len(fp))
            headers["x-kss-meta-unencrypted-content-length"] = str(self.size)
            if self.bucket.connection.enable_crc and not isinstance(fp, utils.FpAdapter):
                fp = utils.FpAdapter(fp)
            resp = self.bucket.connection.make_request(
                method,
                self.bucket.name,
                self.name,
                data=fp,
                headers=headers,
                query_args=query_args,
            )
            if resp and resp.status > 299:
                raise provider.storage_response_error(resp.status, resp.reason, resp.read())
            self.handle_version_headers(resp, force=True)
            self.handle_addl_headers(resp.getheaders())
            if self.bucket.connection.enable_crc:
                self.client_crc = str(fp.crc)
                server_crc = resp.getheader(self.provider.checksum_crc64ecma_header)
                logger.debug('upload key: {0}, bucket: {1}, check crc: [client crc: {2}, server crc: {3}]'
                             .format(self.name, self.bucket.name, self.client_crc, server_crc))
                if not utils.check_crc(self.client_crc, server_crc):
                    raise KS3ClientError("Inconsistent CRC checksum client_crc: %s, server_crc: %s" %
                                         (self.client_crc, resp.getheader(self.provider.checksum_crc64ecma_header)),
                                         resp.headers['x-kss-request-id'])
            return resp
        # keyFile类型不用
        if self.bucket.connection.enable_crc and not isinstance(fp, utils.FpAdapter) and not isinstance(fp, KeyFile):
            fp = utils.FpAdapter(fp)
        resp = self.bucket.connection.make_request(
            method,
            self.bucket.name,
            self.name,
            data=fp,
            headers=headers,
            query_args=query_args
        )
        if resp and resp.status > 299:
            raise provider.storage_response_error(resp.status, resp.reason, resp.read())
        self.handle_version_headers(resp, force=True)
        self.handle_addl_headers(resp.getheaders())
        if self.bucket.connection.enable_crc:
            self.client_crc = str(fp.crc)
            server_crc = resp.getheader(self.provider.checksum_crc64ecma_header)
            logger.debug('upload key: {0}, bucket: {1}, check crc: [client crc: {2}, server crc: {3}]'
                         .format(self.name, self.bucket.name, self.client_crc, server_crc))
            if not utils.check_crc(self.client_crc, server_crc):
                raise KS3ClientError("Inconsistent CRC checksum client_crc: %s, server_crc: %s" %
                                     (self.client_crc, resp.getheader(self.provider.checksum_crc64ecma_header)),
                                     resp.headers['x-kss-request-id'])
        return resp

    def set_contents_from_file(self, fp, method="PUT", headers=None, replace=True,
                               cb=None, num_cb=10, policy=None, md5=None,
                               reduced_redundancy=False, query_args=None,
                               encrypt_key=False, size=None, rewind=False,
                               crypt_context=None, calc_encrypt_md5=True):
        """
        Store an object in S3 using the name of the Key object as the
        key in S3 and the contents of the file pointed to by 'fp' as the
        contents. The data is read from 'fp' from its current position until
        'size' bytes have been read or EOF.
        :type fp: file
        :param fp: the file whose contents to upload

        :type headers: dict
        :param headers: Additional HTTP headers that will be sent with
            the PUT request.

        :type replace: bool
        :param replace: If this parameter is False, the method will
            first check to see if an object exists in the bucket with
            the same key.  If it does, it won't overwrite it.  The
            default value is True which will overwrite the object.

        :type cb: function
        :param cb: a callback function that will be called to report
            progress on the upload.  The callback should accept two
            integer parameters, the first representing the number of
            bytes that have been successfully transmitted to S3 and
            the second representing the size of the to be transmitted
            object.

        :type num_cb: int
        :param num_cb: (optional) If a callback is specified with the
            cb parameter this parameter determines the granularity of
            the callback by defining the maximum number of times the
            callback will be called during the file transfer.

        :type policy: :class:`boto.s3.acl.CannedACLStrings`
        :param policy: A canned ACL policy that will be applied to the
            new key in S3.

        :type md5: A tuple containing the hexdigest version of the MD5
            checksum of the file as the first element and the
            Base64-encoded version of the plain checksum as the second
            element.  This is the same format returned by the
            compute_md5 method.
        :type reduced_redundancy: bool
        :param reduced_redundancy: If True, this will set the storage
            class of the new Key to be REDUCED_REDUNDANCY. The Reduced
            Redundancy Storage (RRS) feature of S3, provides lower
            redundancy at lower storage cost.

        :type encrypt_key: bool
        :param encrypt_key: If True, the new copy of the object will
            be encrypted on the server-side by S3 and will be stored
            in an encrypted form while at rest in S3.

        :type size: int
        :param size: (optional) The Maximum number of bytes to read
            from the file pointer (fp). This is useful when uploading
            a file in multiple parts where you are splitting the file
            up into different ranges to be uploaded. If not specified,
            the default behaviour is to read all bytes from the file
            pointer. Less bytes may be available.

        :type rewind: bool
        :param rewind: (optional) If True, the file pointer (fp) will
            be rewound to the start before any bytes are read from
            it. The default behaviour is False which reads from the
            current position of the file pointer (fp).

        :rtype: int
        :return: The number of bytes written to the key.
        """
        provider = self.bucket.connection.provider
        headers = headers or {}
        if policy:
            headers[provider.acl_header] = policy
        if encrypt_key:
            headers[provider.server_side_encryption_header] = 'AES256'

        if rewind:
            # caller requests reading from beginning of fp.
            fp.seek(0, os.SEEK_SET)
        else:
            # The following seek/tell/seek logic is intended
            # to detect applications using the older interface to
            # set_contents_from_file(), which automatically rewound the
            # file each time the Key was reused. This changed with commit
            # 14ee2d03f4665fe20d19a85286f78d39d924237e, to support uploads
            # split into multiple parts and uploaded in parallel, and at
            # the time of that commit this check was added because otherwise
            # older programs would get a success status and upload an empty
            # object. Unfortuantely, it's very inefficient for fp's implemented
            # by KeyFile (used, for example, by gsutil when copying between
            # providers). So, we skip the check for the KeyFile case.
            # TODO: At some point consider removing this seek/tell/seek
            # logic, after enough time has passed that it's unlikely any
            # programs remain that assume the older auto-rewind interface.
            if not isinstance(fp, KeyFile):
                spos = fp.tell()
                fp.seek(0, os.SEEK_END)
                if fp.tell() == spos:
                    fp.seek(0, os.SEEK_SET)
                    if fp.tell() != spos:
                        # Raise an exception as this is likely a programming
                        # error whereby there is data before the fp but nothing
                        # after it.
                        fp.seek(spos)
                        raise AttributeError('fp is at EOF. Use rewind option '
                                             'or seek() to data start.')
                # seek back to the correct position.
                fp.seek(spos)
        if reduced_redundancy:
            self.storage_class = 'REDUCED_REDUNDANCY'
            if provider.storage_class_header:
                headers[provider.storage_class_header] = self.storage_class
                # TODO - What if provider doesn't support reduced reduncancy?
                # What if different providers provide different classes?
        if hasattr(fp, 'name'):
            self.path = fp.name
        if self.bucket is not None:
            if not md5 and provider.supports_chunked_transfer():
                # defer md5 calculation to on the fly and
                # we don't know anything about size yet.
                chunked_transfer = True
                self.size = None
            else:
                chunked_transfer = False
                if isinstance(fp, KeyFile):
                    # Avoid EOF seek for KeyFile case as it's very inefficient.
                    size = fp.size
                    self.size = size
                    # At present both GCS and S3 use MD5 for the etag for
                    # non-multipart-uploaded objects. If the etag is 32 hex
                    # chars use it as an MD5, to avoid having to read the file
                    # twice while transferring.
                    # keyFile类型，现用于跨region复制，对于源key为分块上传的，没有md5值，所以统一不计算
                    # 通过crc校验一致性，没有crc时，不保证一致性
                    md5 = (None, None)
                    # key = fp.getkey()
                    # if (re.match('^"[a-fA-F0-9]{32}"$', key.etag)):
                    #     etag = key.etag.strip('"')
                    #     md5 = (etag, base64.b64encode(binascii.unhexlify(etag)).decode('utf-8'))
                if not md5:
                    # compute_md5() and also set self.size to actual
                    # size of the bytes read computing the md5.
                    md5 = self.compute_md5(fp, size)
                    # adjust size if required
                    size = self.size
                elif size:
                    self.size = size
                else:
                    # If md5 is provided, still need to size so
                    # calculate based on bytes to end of content
                    spos = fp.tell()
                    fp.seek(0, os.SEEK_END)
                    self.size = fp.tell() - spos
                    fp.seek(spos)
                    size = self.size
                self.md5 = md5[0]
                self.base64md5 = md5[1]

            if self.name is None:
                self.name = self.md5
            if not replace:
                if self.bucket.lookup(self.name):
                    logger.debug('<Key: {0}> in {1} already exists, do not replace.'.format(self.name, self.bucket))
                    return
            try:
                from ks3.encryption import Crypts
            except:
                pass
            if self.bucket.connection.local_encrypt and self.size:
                if not crypt_context:
                    crypt_context = Crypts(self.bucket.connection.key)
                    crypt_context.action_info = "put"
                    crypt_context.calc_md5 = calc_encrypt_md5
                resp = self.send_file(fp, method=method, headers=headers, cb=cb, num_cb=num_cb,
                                      query_args=query_args,
                                      chunked_transfer=chunked_transfer, size=size,
                                      crypt_context=crypt_context)
            else:
                try:
                    resp = self.send_file(fp, method=method, headers=headers, cb=cb, num_cb=num_cb,
                                          query_args=query_args,
                                          chunked_transfer=chunked_transfer, size=size)
                except Exception as e:
                    if isinstance(e, requests.exceptions.ConnectionError) and e.request is not None:
                        e.args = e.args, 'Request URL: {}'.format(e.request.url)
                    raise e
            return ResponseResult(data=resp, status=resp.status, reason=resp.reason,
                          headers=resp.headers)
            # return number of bytes written.
            # return self.size

    def set_contents_from_filename(self, filename, method="PUT", headers=None, replace=True,
                                   cb=None, num_cb=10, policy=None, md5=None,
                                   reduced_redundancy=False,
                                   encrypt_key=False, calc_encrypt_md5=True, query_args=None):
        """
        Store an object in S3 using the name of the Key object as the
        key in S3 and the contents of the file named by 'filename'.
        See set_contents_from_file method for details about the
        parameters.
        """
        with open(filename, 'rb') as fp:
            return self.set_contents_from_file(fp, method, headers, replace, cb,
                                               num_cb, policy, md5,
                                               reduced_redundancy,
                                               encrypt_key=encrypt_key, calc_encrypt_md5=calc_encrypt_md5,
                                               query_args=query_args)

    def append_object_from_filename(self, filename, position, **kwargs):
        return self.set_contents_from_filename(filename, method="POST", query_args="append&position=%d" % position,
                                               **kwargs)

    # 读取大文件慎用——读取全部内容后，再上传，会占用大量内存
    async def upload_file_async(self, filename, method="PUT", headers=None, replace=True, cb=None, num_cb=10, policy=None,
                                md5=None,
                                reduced_redundancy=False, encrypt_key=False, calc_encrypt_md5=True,
                                query_args=None):
        """
        set_contents_from_filename的异步版本
        """
        import asyncio
        from concurrent.futures import ThreadPoolExecutor

        data = await _read_file_async(filename)
        loop = asyncio.get_running_loop()
        # 使用线程池执行同步函数，实现异步效果
        with ThreadPoolExecutor() as pool:
            r = await loop.run_in_executor(pool, lambda: self.set_contents_from_string(
                data, method, headers, replace, cb,  num_cb, policy, md5, reduced_redundancy, encrypt_key=encrypt_key,
                calc_encrypt_md5=calc_encrypt_md5, query_args=query_args))
        return r

    def set_contents_from_string(self, string_data, method="PUT", headers=None, replace=True,
                                 cb=None, num_cb=10, policy=None, md5=None,
                                 reduced_redundancy=False,
                                 encrypt_key=False, calc_encrypt_md5=True, query_args=None):
        """
        Store an object in S3 using the name of the Key object as the
        key in S3 and the string 's' as the contents.
        See set_contents_from_file method for details about the
        parameters.
        """
        if not isinstance(string_data, bytes):
            string_data = string_data.encode('utf-8')
        fp = BytesIO(string_data)
        r = self.set_contents_from_file(fp, method, headers, replace, cb, num_cb,
                                        policy, md5, reduced_redundancy,
                                        encrypt_key=encrypt_key, calc_encrypt_md5=calc_encrypt_md5,
                                        query_args=query_args)
        fp.close()
        return r

    def append_object_from_string(self, string_data, position, **kwargs):
        return self.set_contents_from_string(string_data, method="POST", query_args="append&position=%d" % position,
                                             **kwargs)

    def get_file(self, fp, headers=None, cb=None, num_cb=10,
                 torrent=False, version_id=None, override_num_retries=None,
                 response_headers=None):
        """
        Retrieves a file from an S3 Key
        """
        return self._get_file_internal(fp, headers=headers, cb=cb, num_cb=num_cb,
                                       torrent=torrent, version_id=version_id,
                                       override_num_retries=override_num_retries,
                                       response_headers=response_headers,
                                       hash_algs=None,
                                       query_args=None)

    def _get_file_internal(self, fp, headers=None, cb=None, num_cb=10,
                           torrent=False, version_id=None, override_num_retries=None,
                           response_headers=None, hash_algs=None, query_args=None):
        if headers is None:
            headers = {}
        save_debug = self.bucket.connection.debug
        if self.bucket.connection.debug == 1:
            self.bucket.connection.debug = 0

        query_args = query_args or []
        if torrent:
            query_args.append('torrent')

        if hash_algs is None and not torrent:
            hash_algs = {'md5': md5}
        digesters = dict((alg, hash_algs[alg]()) for alg in hash_algs or {})

        # If a version_id is passed in, use that.  If not, check to see
        # if the Key object has an explicit version_id and, if so, use that.
        # Otherwise, don't pass a version_id query param.
        if version_id is None:
            version_id = self.version_id
        if version_id:
            query_args.append('versionId=%s' % version_id)
        if response_headers:
            for key in response_headers:
                query_args.append('%s=%s' % (
                    key, parse.quote(response_headers[key])))
        query_args = '&'.join(query_args)

        if self.bucket.connection.local_encrypt:
            start_offset = 0
            if self.size is None:
                response = self.bucket.connection.make_request('HEAD', self.bucket.name, self.name)
                if response.status < 199 or response.status > 299:
                    raise self.bucket.connection.provider.storage_response_error(response.status,
                                                                                 response.reason, None)
            end_offset = self.size - 1
            skip_bytes = 0
            cut_bytes = 0
            decrypted_total_part = None
        if 'range' in headers and self.bucket.connection.local_encrypt:
            range_value = headers['range'].replace("bytes=", "").strip()
            start_end = range_value.split('-')
            if len(start_end) == 2:
                desired_start = int(start_end[0]) if start_end[0] else 0
                desired_end = int(start_end[1]) if start_end[1] else self.size
                # Takes the position of the leftmost desired byte of a user specified range and calculates the
                # position of the start of the previous cipher block, or set 0 if the leftmost byte is in
                # the first cipher block.
                start_offset = max(desired_start - (desired_start % 16) - 16, 0)
                # Takes the position of the rightmost desired byte of a user specified range and calculates the
                # position of the end of the following cipher block.
                end_offset = min(desired_end + (16 - (desired_end % 16)) + 16 - 1, end_offset)

                headers['range'] = 'bytes={0}-{1}'.format(str(start_offset), str(end_offset))
                # When local encryption enabled, the value of the range header will be expanded,
                # resulting in more data than expected after decryption.
                # we need to calculate the length that needs to be skipped at the beginning
                # and the length that needs to be truncated at the end based on the user desired range.
                skip_bytes = max(desired_start - start_offset, 0)
                cut_bytes = end_offset - 16 - desired_end

                decrypted_size = end_offset - start_offset + 1 - 16
                decrypted_total_part = math.ceil(float(decrypted_size) / self.BufferSize)

        resp_result = self.open('r', headers, query_args=query_args,
                                override_num_retries=override_num_retries)

        data_len = 0
        if cb:
            if self.size is None:
                cb_size = 0
            else:
                cb_size = self.size
            if self.size is None and num_cb != -1:
                # If size is not available due to chunked transfer for example,
                # we'll call the cb for every 1MB of data transferred.
                cb_count = (1024 * 1024) / self.BufferSize
            elif num_cb > 1:
                cb_count = int(math.ceil(cb_size / self.BufferSize / (num_cb - 1.0)))
            elif num_cb < 0:
                cb_count = -1
            else:
                cb_count = 0
            i = 0
            cb(data_len, cb_size)
        try:
            counter = 1
            last_iv = ""
            # When end_offset + 1 === self.size, calculate the data size through start_offset and end_offset
            # In order to adapt to the situation when the range header is provided
            if self.bucket.connection.local_encrypt:
                if end_offset + 1 == self.size:
                    # adapt to the case where the range header (bytes=start-end) end + 1 == self.size
                    total_part = math.ceil(float(end_offset - start_offset + 1) / self.BufferSize)
                else:
                    # the case that the range header is not provided
                    total_part = math.ceil(float(self.size) / self.BufferSize)
            for byte in self:
                if self.bucket.connection.local_encrypt:
                    provider = self.bucket.connection.provider
                    user_key = self.bucket.connection.key
                    crypt_handler = Crypts(user_key)
                    if counter == 1:
                        # For first block, drop first 16 bytes(the subjoin iv).
                        local_iv = byte[:crypt_handler.block_size]
                        byte = byte[crypt_handler.block_size:]
                        server_iv = self.user_meta[provider.metadata_prefix + "iv"]
                        server_iv = base64.b64decode(server_iv)
                        if start_offset == 0 and server_iv and local_iv != server_iv:
                            raise ValueError("decryption error:file.iv not equel server.iv")
                        user_iv = local_iv
                    else:
                        user_iv = last_iv
                    last_iv = byte[-crypt_handler.block_size:]
                    # adapt to the case where the range header (bytes=start-end) end + 1 < self.size
                    # In order to adapt to the situation when the range header is provided
                    # The following two conditions need to be met
                    if end_offset + 1 == self.size and counter == total_part:
                        # Special process of the last part with check code appending to it's end.
                        full_content = crypt_handler.decrypt(byte, user_iv)
                        pad_content_char = full_content[-1]
                        if isinstance(pad_content_char, int):
                            pad_content_char = chr(pad_content_char)
                        for key in crypt_handler.pad_dict:
                            if crypt_handler.pad_dict[key] == pad_content_char:
                                pad_content_char = key
                        decrypt = full_content[:-int(pad_content_char)]
                    else:
                        decrypt = crypt_handler.decrypt(byte, user_iv)
                    byte = decrypt
                    if counter == 1:
                        byte = byte[skip_bytes:]
                    if decrypted_total_part is not None and counter == decrypted_total_part and cut_bytes > 0:
                        byte = byte[:-cut_bytes]
                    counter += 1
                if not isinstance(byte, bytes):
                    byte = byte.encode()
                fp.write(byte)
                data_len += len(byte)
                for alg in digesters:
                    digesters[alg].update(byte)
                if cb:
                    if cb_size > 0 and data_len >= cb_size:
                        break
                    i += 1
                    if i == cb_count or cb_count == -1:
                        cb(data_len, cb_size)
                        i = 0
        except IOError as e:
            if e.errno == errno.ENOSPC:
                raise StorageDataError('Out of space for destination file '
                                       '%s' % fp.name)
            raise
        if cb and (cb_count <= 1 or i > 0) and data_len > 0:
            cb(data_len, cb_size)
        for alg in digesters:
            self.local_hashes[alg] = digesters[alg].digest()
        if self._size is None and not torrent and "Range" not in headers:
            self.size = data_len
        self.close()
        self.bucket.connection.debug = save_debug

        return resp_result

    def open_read(self, headers=None, query_args='',
                  override_num_retries=None, response_headers=None):
        """
        Open this key for reading
        """
        if self.resp is None:
            self.mode = 'r'

            provider = self.bucket.connection.provider
            self.resp = self.bucket.connection.make_request(
                'GET', self.bucket.name, self.name, headers=headers,
                query_args=query_args)
            if self.resp.status < 199 or self.resp.status > 299:
                body = self.resp.read()
                raise provider.storage_response_error(self.resp.status,
                                                      self.resp.reason, body)
            response_headers = self.resp.msg
            # self.metadata = boto.utils.get_aws_metadata(response_headers,
            #                                            provider)
            for name, value in list(response_headers.items()):
                # To get correct size for Range GETs, use Content-Range
                # header if one was returned. If not, use Content-Length
                # header.
                if (name.lower() == 'content-length' and
                        'Content-Range' not in response_headers):
                    self.size = int(value)
                elif name.lower() == 'content-range':
                    end_range = re.sub('.*/(.*)', '\\1', value)
                    self.size = int(end_range)
                elif name.lower() in Key.base_fields:
                    self.__dict__[name.lower().replace('-', '_')] = value
            self.handle_version_headers(self.resp)
            self.handle_encryption_headers(self.resp)
            self.handle_restore_headers(self.resp)
            self.handle_addl_headers(self.resp.getheaders())
            self.handle_user_metas(self.resp)
            self.handle_storage_class(self.resp)
            self.handle_checksum_crc64ecma(self.resp)
        return ResponseResult(data=None, status=self.resp.status, reason=self.resp.reason,
                              headers=self.resp.headers)

    def open_write(self, headers=None, override_num_retries=None):
        """
        Open this key for writing.
        Not yet implemented

        :type headers: dict
        :param headers: Headers to pass in the write request

        :type override_num_retries: int
        :param override_num_retries: If not None will override configured
            num_retries parameter for underlying PUT.
        """
        raise KS3ClientError('Not Implemented')

    def open(self, mode='r', headers=None, query_args=None,
             override_num_retries=None):
        if mode == 'r':
            self.mode = 'r'
            return self.open_read(headers=headers, query_args=query_args,
                                  override_num_retries=override_num_retries)
        elif mode == 'w':
            self.mode = 'w'
            self.open_write(headers=headers,
                            override_num_retries=override_num_retries)
        else:
            raise KS3ClientError('Invalid mode: %s' % mode)

    closed = False

    def close(self, fast=False):
        """
        Close this key.

        :type fast: bool
        :param fast: True if you want the connection to be closed without first
        reading the content. This should only be used in cases where subsequent
        calls don't need to return the content from the open HTTP connection.
        Note: As explained at
        http://docs.python.org/2/library/httplib.html#httplib.HTTPConnection.getresponse,
        callers must read the whole response before sending a new request to the
        server. Calling Key.close(fast=True) and making a subsequent request to
        the server will work because boto will get an httplib exception and
        close/reopen the connection.

        """
        if self.resp and not fast:
            self.resp.read()
        self.resp = None
        self.mode = None
        self.closed = True

    def get_contents_to_file(self, fp, byte_range=None, headers=None,
                             cb=None, num_cb=10,
                             torrent=False,
                             version_id=None,
                             res_download_handler=None,
                             response_headers=None):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Write the contents of the object to the file pointed
        to by 'fp'.

        :param byte_range: 指定下载范围。参见: ref:`byte_range`
        """

        range_string = _make_range_string(byte_range)
        if range_string:
            if not headers:
                headers = {}
            headers['range'] = range_string

        if self.bucket is not None:
            if res_download_handler:
                return res_download_handler.get_file(self, fp, headers, cb, num_cb,
                                                     torrent=torrent,
                                                     version_id=version_id)
            else:
                return self.get_file(fp, headers, cb, num_cb, torrent=torrent,
                                     version_id=version_id,
                                     response_headers=response_headers)

    def get_contents_to_filename(self, filename, byte_range=None, headers=None,
                                 cb=None, num_cb=10,
                                 torrent=False,
                                 version_id=None,
                                 res_download_handler=None,
                                 response_headers=None):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Store contents of the object to a file named by 'filename'.
        See get_contents_to_file method for details about the
        parameters.

        :param byte_range: 指定下载范围。参见: ref:`byte_range`
        """

        range_string = _make_range_string(byte_range)
        if range_string:
            if not headers:
                headers = {}
            headers['range'] = range_string

        directory_path = os.path.dirname(filename)
        base_name = os.path.basename(filename)
        try:
            with tempfile.NamedTemporaryFile(
                    mode='wb', dir=directory_path, prefix=base_name + '.', suffix='.ks3temp', delete=False) as tmp_fp:
                tmp_name = tmp_fp.name
                resp_result = self.get_contents_to_file(tmp_fp, byte_range, headers, cb, num_cb,
                                          torrent=torrent,
                                          version_id=version_id,
                                          res_download_handler=res_download_handler,
                                          response_headers=response_headers)
            if os.path.exists(filename):
                os.remove(filename)
            os.rename(tmp_name, filename)
        except Exception:
            os.remove(tmp_name)
            raise
        # if last_modified date was sent from s3, try to set file's timestamp
        if self.last_modified is not None:
            try:
                modified_tuple = email.utils.parsedate_tz(self.last_modified)
                modified_stamp = int(email.utils.mktime_tz(modified_tuple))
                os.utime(filename, (modified_stamp, modified_stamp))
            except:
                pass
        return resp_result

    async def download_file_async(self, filename, byte_range=None, headers=None,
                                  cb=None, num_cb=10,
                                  torrent=False,
                                  version_id=None,
                                  response_headers=None):

        import asyncio
        from concurrent.futures import ThreadPoolExecutor

        loop = asyncio.get_running_loop()
        # 使用线程池执行同步函数，实现异步效果
        with ThreadPoolExecutor() as pool:
            ret = await loop.run_in_executor(pool, lambda: self.get_contents_as_obj_with_string(
                byte_range, headers, cb, num_cb, torrent=torrent, version_id=version_id,
                response_headers=response_headers))
        await _write_file_async(filename, ret.data)
        return ret

    def get_contents_as_string(self, byte_range=None, headers=None,
                               cb=None, num_cb=10,
                               torrent=False,
                               version_id=None,
                               response_headers=None, encoding=None):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Return the contents of the object as a string.
        See get_contents_to_file method for details about the
        parameters.

        :param byte_range: 指定下载范围。参见: ref:`byte_range`
        """

        fp = BytesIO()
        self.get_contents_to_file(fp, byte_range, headers, cb, num_cb, torrent=torrent,
                                  version_id=version_id,
                                  response_headers=response_headers)
        value = fp.getvalue()
        if encoding is not None:
            value = value.decode(encoding)
        return value

    def get_contents_as_obj_with_string(self, byte_range=None, headers=None,
                               cb=None, num_cb=10,
                               torrent=False,
                               version_id=None,
                               response_headers=None, encoding=None):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Return the contents of the object as a string.
        See get_contents_to_file method for details about the
        parameters.

        :param byte_range: 指定下载范围。参见: ref:`byte_range`
        """

        fp = BytesIO()
        resp = self.get_contents_to_file(fp, byte_range, headers, cb, num_cb, torrent=torrent,
                                  version_id=version_id,
                                  response_headers=response_headers)
        value = fp.getvalue()
        if encoding is not None:
            value = value.decode(encoding)
        resp.data = value
        return resp

    def restore_object(self, days=None, headers=None):
        return self.bucket.restore_object(self.name, days=days, headers=headers)

    def set_object_tagging(self, tagging_set, version_id=None, headers=None):
        object_tagging = ObjectTagging(tagging_set)
        if headers is None:
            headers = {}
        headers['Content-Type'] = 'text/xml'
        object_tagging_xml = object_tagging.to_xml()
        if not isinstance(object_tagging_xml, bytes):
            object_tagging_xml = object_tagging_xml.encode('utf-8')

        query_args = 'tagging'
        if version_id is not None:
            query_args = query_args + '&versionId=' + version_id
        response = self.bucket.connection.make_request('PUT', self.bucket.name, self.name, data=object_tagging_xml,
                                                       query_args=query_args, headers=headers)

        body = response.read()
        if response.status == 200:
            return ResponseResult(True, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_object_tagging(self, version_id=None, headers=None):
        query_args = 'tagging'
        if version_id is not None:
            query_args = query_args + '&versionId=' + version_id
        response = self.bucket.connection.make_request('GET', self.bucket.name, self.name,
                                                       query_args=query_args, headers=headers)
        body = response.read()
        if response.status == 200:
            objectTagging = ObjectTagging(status=response.status, reason=response.reason, headers=response.headers,
                                          raw_body=body)
            h = handler.XmlHandler(objectTagging, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return objectTagging
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_object_tagging(self, version_id=None, headers=None):
        query_args = 'tagging'
        if version_id is not None:
            query_args = query_args + '&versionId=' + version_id
        response = self.bucket.connection.make_request('DELETE', self.bucket.name, self.name,
                                                       query_args=query_args, headers=headers)
        body = response.read()
        if response.status == 204:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def fetch_object(self, source_url=None, callback_url=None, headers=None):
        return self.bucket.fetch_object(self.name, source_url=source_url, callback_url=callback_url, headers=headers)

    def recover_object(self, overwrite=None, retention_id=None, headers=None):
        if headers is None:
            headers = {}
        if overwrite is not None:
            headers[self.provider.retention_overwrite_header] = 'true' if overwrite else 'false'
        if retention_id is not None:
            headers[self.provider.retention_id_header] = retention_id
        query_args = 'recover'

        response = self.bucket.connection.make_request('POST', self.bucket.name, self.name,
                                                       query_args=query_args, headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def clear_object(self, retention_id=None, headers=None):
        if headers is None:
            headers = {}
        if retention_id is not None:
            headers[self.provider.retention_id_header] = retention_id
        query_args = 'clear'
        response = self.bucket.connection.make_request('DELETE', self.bucket.name, self.name,
                                                       query_args=query_args, headers=headers)
        body = response.read()
        if response.status == 200:
            return ResponseResult(None, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def upload_file(
            self,
            filename,
            part_size=DEFAULT_PART_SIZE,
            threads_num=3,
            resumable=False,
            resumable_filename=None,
            headers=None,
    ):
        real_threads_num = threads_num
        if self.bucket.connection.local_encrypt:
            real_threads_num = 1
            logger.debug("key_name={0}, local_encrypt={1}, set threads_num to {2}"
                         .format(self.name, self.bucket.connection.local_encrypt, real_threads_num))
        executor = BlockThreadPoolExecutor(
            max_workers=real_threads_num,
            thread_name_prefix='upload-task',
        )
        return UploadTask(
            self,
            self.bucket,
            filename,
            executor,
            part_size=part_size,
            resumable=resumable,
            resumable_filename=resumable_filename,
        ).upload(headers)

    def download_file(
            self,
            filename,
            part_size=DEFAULT_PART_SIZE,
            threads_num=3,
            resumable=False,
            resumable_filename=None,
            headers=None,
    ):
        executor = BlockThreadPoolExecutor(
            max_workers=threads_num,
            thread_name_prefix='download-task',
        )
        return DownloadTask(
            self,
            self.bucket,
            filename,
            executor,
            part_size=part_size,
            resumable=resumable,
            resumable_filename=resumable_filename,
        ).download(headers)

    def copy_file(
            self,
            src_key,
            part_size=DEFAULT_PART_SIZE,
            threads_num=3,
            resumable=False,
            resumable_filename=None,
            headers=None,
            copy_src_metadata=True,
            copy_src_tagging=True,
    ):
        executor = BlockThreadPoolExecutor(
            max_workers=threads_num,
            thread_name_prefix='copy-task',
        )
        return CopyTask(
            src_key,
            self,
            executor,
            part_size=part_size,
            resumable=resumable,
            resumable_filename=resumable_filename,
            copy_src_metadata=copy_src_metadata,
            copy_src_tagging=copy_src_tagging,
        ).copy(headers)

    def across_region_copy_file(
            self,
            src_key,
            part_size=DEFAULT_PART_SIZE,
            threads_num=3,
            resumable=False,
            resumable_filename=None,
            headers=None,
            copy_src_metadata=True,
            copy_src_tagging=True,
    ):
        executor = BlockThreadPoolExecutor(
            max_workers=threads_num,
            thread_name_prefix='across-region-copy-task',
        )
        return AcrossRegionCopyTask(
            src_key,
            self,
            executor,
            part_size=part_size,
            resumable=resumable,
            resumable_filename=resumable_filename,
            copy_src_metadata=copy_src_metadata,
            copy_src_tagging=copy_src_tagging,
        ).copy(headers)

    def set_object_migration(self, src_bucket_name, src_key_name, storage_class=None, headers=None):
        if headers is None:
            headers = {}
        migration_source = '/{0}/{1}'.format(src_bucket_name, url_encode(src_key_name)).replace('//', '/%2F')
        headers[self.provider.migration_source_header] = migration_source
        if storage_class is not None:
            headers[self.provider.storage_class_header] = storage_class

        response = self.bucket.connection.make_request('PUT', self.bucket.name, self.name, headers=headers,
                                                       query_args='migration')
        body = response.read()
        if response.status == 200:
            result = MigrationResult(status=response.status, reason=response.reason, headers=response.headers)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_object_migration(self, headers=None):
        if headers is None:
            headers = {}
        query_args = 'migration'
        response = self.bucket.connection.make_request('GET', self.bucket.name, self.name, headers=headers,
                                                       query_args=query_args)
        body = response.read()
        if response.status == 200:
            result = MigrationConfiguration(status=response.status, reason=response.reason, headers=response.headers)
            h = handler.XmlHandler(result, self)
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            xml.sax.parseString(body, h)
            return result
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def handle_version_headers(self, resp, force=False):
        provider = self.bucket.connection.provider
        # If the Key object already has a version_id attribute value, it
        # means that it represents an explicit version and the user is
        # doing a get_contents_*(version_id=<foo>) to retrieve another
        # version of the Key.  In that case, we don't really want to
        # overwrite the version_id in this Key object.  Comprende?
        if self.version_id is None or force:
            self.version_id = resp.getheader(provider.version_id, None)
        self.source_version_id = resp.getheader(provider.copy_source_version_id,
                                                None)
        if resp.getheader(provider.delete_marker, 'false') == 'true':
            self.delete_marker = True
        else:
            self.delete_marker = False

    def handle_encryption_headers(self, resp):
        provider = self.bucket.connection.provider
        if provider.server_side_encryption_header:
            self.encrypted = resp.getheader(
                provider.server_side_encryption_header, None)
        else:
            self.encrypted = None

    def handle_restore_headers(self, response):
        provider = self.bucket.connection.provider
        header = response.getheader(provider.restore_header)
        if header is None:
            return
        parts = header.split(',', 1)
        for part in parts:
            key, val = [i.strip() for i in part.split('=')]
            val = val.replace('"', '')
            if key == 'ongoing-request':
                self.ongoing_restore = True if val.lower() == 'true' else False
            elif key == 'expiry-date':
                self.expiry_date = val

    def handle_addl_headers(self, headers):
        """
        Used by Key subclasses to do additional, provider-specific
        processing of response headers. No-op for this base class.
        """
        pass

    def handle_user_metas(self, response):
        """
        If there are user set custom metas, this function will put them into a dict.
        """
        # print response.msg
        provider = self.bucket.connection.provider
        user_key_reg = provider.metadata_prefix + ".*"
        for header_key, header_value in list(response.msg.items()):
            reg_match = re.match(user_key_reg, header_key)
            if reg_match:
                self.user_meta[header_key] = header_value

    def handle_storage_class(self, response):
        provider = self.bucket.connection.provider
        if provider:
            sc = response.getheader(provider.storage_class_header)
            if sc:
                self.storage_class = sc
            else:
                self.storage_class = "STANDARD"

    def compute_md5(self, fp, size=None):
        hex_digest, b64_digest, data_size = compute_md5(fp, size=size)
        self.size = data_size
        return (hex_digest, b64_digest)

    def compute_encrypted_md5(self, fp):
        hex_digest, b64_digest = compute_encrypted_md5(fp)
        return b64_digest

    def handle_tagging_count(self, response):
        provider = self.bucket.connection.provider
        if provider:
            sc = response.getheader(provider.tagging_count_header)
            if sc:
                self.tagging_count = sc

    def handle_object_type(self, response):
        provider = self.bucket.connection.provider
        if provider:
            sc = response.getheader(provider.object_type_header)
            if sc:
                self.object_type = sc

    def handle_next_position(self, response):
        provider = self.bucket.connection.provider
        if provider:
            sc = response.getheader(provider.append_next_position_header)
            if sc:
                self.append_next_position = sc

    def handle_checksum_crc64ecma(self, response):
        provider = self.bucket.connection.provider
        if provider:
            sc = response.getheader(provider.checksum_crc64ecma_header)
            if sc:
                self.server_crc = sc


def _make_range_string(range):
    if range is None:
        return ''

    start = range[0]
    last = range[1]

    if start is None and last is None:
        return ''

    return 'bytes=' + _range(start, last)


def _range(start, last):
    def to_str(pos):
        if pos is None:
            return ''
        else:
            return str(pos)

    return to_str(start) + '-' + to_str(last)


async def _read_file_async(filename):
    async with aiofiles.open(filename, mode='rb') as file:
        return await file.read()


async def _write_file_async(filename, data):
    async with aiofiles.open(filename, mode='wb') as file:
        return await file.write(data)



# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
from ks3.responseResult import ResponseMetadata
from ks3.user import User

CannedACLStrings = ['private', 'public-read',
                    'public-read-write', 'authenticated-read',
                    'bucket-owner-read', 'bucket-owner-full-control',
                    'log-delivery-write']


# AccessControlPolicy
class Policy(object):

    def __init__(self, parent=None, *args, **kwargs):
        self.parent = parent
        self.namespace = None
        self.acl = None

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        grants = []
        for g in self.acl.grants:
            if g.id == self.owner.id:
                grants.append("%s (owner) = %s" % (g.display_name, g.permission))
            else:
                if g.type == 'CanonicalUser':
                    u = g.display_name
                elif g.type == 'Group':
                    u = g.uri
                else:
                    u = g.email_address
                grants.append("%s = %s" % (u, g.permission))
        return "<Policy: %s>" % ", ".join(grants)

    def startElement(self, name, attrs, connection):
        if name == 'AccessControlPolicy':
            self.namespace = attrs.get('xmlns', None)
            return None
        if name == 'Owner':
            self.owner = User(self)
            return self.owner
        elif name == 'AccessControlList':
            self.acl = ACL(self)
            return self.acl
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Owner':
            pass
        elif name == 'AccessControlList':
            pass
        else:
            setattr(self, name, value)

    def to_xml(self):
        if self.namespace is not None:
            s = '<AccessControlPolicy xmlns="{0}">'.format(self.namespace)
        else:
            s = '<AccessControlPolicy xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
        s += self.owner.to_xml()
        s += self.acl.to_xml()
        s += '</AccessControlPolicy>'
        return s


class ACL(object):

    def __init__(self, policy=None):
        self.policy = policy
        self.grants = []

    def add_grant(self, grant):
        self.grants.append(grant)

    def add_email_grant(self, permission, email_address):
        grant = Grant(permission=permission, type='AmazonCustomerByEmail',
                      email_address=email_address)
        self.grants.append(grant)

    def add_user_grant(self, permission, user_id, display_name=None):
        grant = Grant(permission=permission, type='CanonicalUser', id=user_id, display_name=display_name)
        self.grants.append(grant)

    def startElement(self, name, attrs, connection):
        if name == 'Grant':
            self.grants.append(Grant(self))
            return self.grants[-1]
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Grant':
            pass
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<AccessControlList>'
        for grant in self.grants:
            s += grant.to_xml()
        s += '</AccessControlList>'
        return s


class Grant(object):
    NameSpace = 'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'

    def __init__(self, permission=None, type=None, id=None,
                 display_name=None, uri=None, email_address=None):
        self.permission = permission
        self.id = id
        self.display_name = display_name
        self.uri = uri
        self.email_address = email_address
        self.type = type

    def startElement(self, name, attrs, connection):
        if name == 'Grantee':
            self.type = attrs['xsi:type']
        return None

    def endElement(self, name, value, connection):
        if name == 'ID':
            self.id = value
        elif name == 'DisplayName':
            self.display_name = value
        elif name == 'URI':
            self.uri = value
        elif name == 'EmailAddress':
            self.email_address = value
        elif name == 'Grantee':
            pass
        elif name == 'Permission':
            self.permission = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<Grant>'
        s += '<Grantee %s xsi:type="%s">' % (self.NameSpace, self.type)
        if self.type == 'CanonicalUser':
            s += '<ID>%s</ID>' % self.id
            s += '<DisplayName>%s</DisplayName>' % self.display_name
        elif self.type == 'Group':
            s += '<URI>%s</URI>' % self.uri
        else:
            s += '<EmailAddress>%s</EmailAddress>' % self.email_address
        s += '</Grantee>'
        s += '<Permission>%s</Permission>' % self.permission
        s += '</Grant>'
        return s



# coding=utf-8
# 多线程分块上传和下载样例

from ks3.connection import Connection
# 需要安装filechunkio
from filechunkio import FileChunkIO
import threading
from queue import Queue
import datetime
import os, math

ak = 'YOUR_ACCESS_KEY'
sk = 'YOUR_SECRET_KEY'
conn = Connection(ak, sk, host="ks3-cn-beijing.ksyuncs.com")
test_bucket = 'sdktestt'
test_key = 'test_key'


def single_multipart_upload(bucket, file_path, test_key):
    print('single_multipart_upload')
    source_path = file_path
    source_size = os.stat(source_path).st_size

    print('multi-upload start:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])
    mp = bucket.initiate_multipart_upload(test_key, policy="public-read-write")
    # print(mp.id)

    chunk_size = 5242880
    chunk_count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

    try:
        for i in range(chunk_count):
            offset = chunk_size * i
            bytes = min(chunk_size, source_size - offset)
            with FileChunkIO(source_path, 'r', offset=offset, bytes=bytes) as fp:
                mp.upload_part_from_file(fp, part_num=i + 1)
        ret = mp.complete_upload()
        if ret and ret.status == 200:
            print('multi-upload succeed:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])
        else:
            print('multi-upload failed:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])
    except:
        print('multi-upload failed:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])


class Chunk:
    num = 0
    offset = 0
    len = 0

    def __init__(self, n, o, l):
        self.num = n
        self.offset = o
        self.len = l


def init_queue(file_size, chunk_size):
    chunkcnt = int(math.ceil(file_size*1.0/chunk_size))
    #print('chunk count:', chunkcnt)
    q = Queue(maxsize = chunkcnt)
    for i in range(0,chunkcnt):
        offset = chunk_size*i
        len = min(chunk_size, file_size-offset)
        c = Chunk(i+1, offset, len)
        q.put(c)
    return q


def upload_chunk(file_path, mp, q):
    while (not q.empty()):
        chunk = q.get()
        fp = FileChunkIO(file_path, 'r', offset=chunk.offset, bytes=chunk.len)
        # print('file%s chunk_%s start %s \n' % (file_path, chunk.num, datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3]))
        mp.upload_part_from_file(fp, part_num=chunk.num)
        # print('file%s chunk_%s end %s \n' % (file_path, chunk.num, datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3]))
        fp.close()
        q.task_done()

# thread_cnt 为线程数
def threading_multipart_upload(bucket, file_path, key_name, thread_cnt=10):
    print('multi-upload ', key_name, ' start:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])
    file_size = os.stat(file_path).st_size
    mp = bucket.initiate_multipart_upload(key_name)
    chunk_size = 5242880
    q = init_queue(file_size, chunk_size)
    if thread_cnt is None:
        thread_cnt = q.qsize() - 1
    for i in range(0, thread_cnt):
        t = threading.Thread(target=upload_chunk, args=(file_path, mp, q))
        t.setDaemon(True)
        t.start()
    q.join()
    try:
        ret = mp.complete_upload()
        if ret and ret.status == 200:
            print('multi-upload succeed:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])
        else:
            print('multi-upload failed:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])
    except:
        print('multi-upload failed:', datetime.datetime.now().strftime("%H:%M:%S.%f'")[:-3])


if __name__=="__main__":
    bucket = conn.get_bucket(test_bucket)
    # 单线程执行
    # print('单线程执行')
    # single_multipart_upload(bucket, file_path, test_key)
    # 多线程执行
    print('多线程执行')
    threading_multipart_upload(bucket, file_path, test_key, thread_cnt=10)



# coding=utf-8
import math
import os
import time

import ks3.multipart
from ks3 import utils
from ks3.connection import Connection

ak = os.getenv('KS3_TEST_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
sk = os.getenv('KS3_TEST_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
bucket_name = os.getenv('KS3_TEST_BUCKET', '<KS3_TEST_BUCKET>')
endpoint = os.getenv('KS3_TEST_ENDPOINT', 'ks3-cn-shanghai-internal.ksyuncs.com')

key_name = '<your_key_name>'
file_path = '<your_file_path>'

# 默认开启自动校验crc
# 关闭需要 conn.enable_crc = False 或 Connection(ak, sk, host=endpoint, enable_crc=False)
conn = Connection(ak, sk, host=endpoint)
bucket = conn.get_bucket(bucket_name)


def test_uploadObjectFromFile_checkingCrc64():
    # 自动crc校验
    k = bucket.new_key(key_name)
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200


def test_uploadObjectFromString_checkingCrc64():
    # 自动crc校验
    k = bucket.new_key(key_name)
    myContent = 'some string'
    ret = k.set_contents_from_string(myContent)
    assert ret.status == 200


def test_downloadObject_checkingCrc64():
    k = bucket.get_key(key_name)

    data = k.read(300)
    while data:
        s = data.decode()
        print('bytes decoded:', s)
        time.sleep(5)
        data = k.read(300)

    assert k.server_crc == k.client_crc


def test_downloadObjectToString_checkingCrc64():
    k = bucket.get_key(key_name)
    s = k.get_contents_as_string().decode()
    client_crc = utils.compute_data_crc64(s)
    assert client_crc == k.server_crc


def test_headObject_checkingCrc64():
    k = bucket.new_key(key_name)
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200

    k = bucket.get_key(key_name)
    client_crc = utils.compute_file_crc64(file_path)
    assert client_crc == k.server_crc


def test_copyObject_checkingCrc64():
    k = bucket.new_key(key_name)
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200

    ret = bucket.copy_key("copy_test", bucket_name, key_name)
    client_crc = utils.compute_file_crc64(file_path)
    assert client_crc == ret.server_crc


def test_multi_upload_checkingCrc64():
    # 自动校验crc
    from filechunkio import FileChunkIO

    source_size = os.stat(file_path).st_size

    mp = bucket.initiate_multipart_upload("test_multi")

    chunk_size = 100 * 1024
    count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

    # 默认自动校验每个分片的crc，和合并后的crc
    for i in range(count):
        offset = chunk_size * i
        read_size = min(chunk_size, source_size - offset)
        with FileChunkIO(file_path, 'r', offset=offset, bytes=read_size) as fp:
            # 逐个上传分片
            mp.upload_part_from_file(fp, part_num=i + 1)
    ret = mp.complete_upload()
    assert ret and ret.status == 200


def test_multi_copy_checkingCrc64():
    k = bucket.new_key("test_multi")
    ret = k.set_contents_from_filename(file_path)
    assert ret.status == 200

    copy_key = 'test_multi-copy'
    source_size = os.stat(file_path).st_size

    mp = bucket.initiate_multipart_upload(copy_key)

    chunk_size = 100 * 1024
    count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

    help_list = []

    for i in range(count):
        start = chunk_size * i
        end = min(start + chunk_size - 1, source_size - 1)
        ret = mp.copy_part_from_key(bucket_name, "test_multi", i + 1, start, end)
        client_crc = utils.compute_file_crc64(file_path, start, end)
        help_list.append(ks3.multipart.PartInfo(end - start + 1, client_crc))
        assert client_crc == ret.server_crc
    ret = mp.complete_upload()
    parts_crc = 0
    crc_obj = utils.Crc64(0)
    for p in help_list:
        parts_crc = crc_obj.combine(parts_crc, int(p.part_crc), p.size)
    assert str(parts_crc) == ret.ChecksumCRC64ECMA


########## 手动计算crc，通过请求头校验的方式 ##########
# ks3服务端也提供了crc64校验的请求头，如果不开启自动校验crc的功能，可以通过传入请求头的方式让服务端进行crc的比对

def test_uploadObjectFromFile_checkingCrc64ByHeader():
    k = bucket.new_key(key_name)
    conn.enable_crc = False
    local_crc = utils.compute_file_crc64(file_path)
    crc_header = conn.provider.checksum_crc64ecma_header
    k.set_contents_from_filename(file_path, headers={crc_header: local_crc})


def test_uploadObjectFromString_checkingCrc64ByHeader():
    k = bucket.new_key(key_name)
    myContent = 'some string'
    conn.enable_crc = False
    local_crc = utils.compute_data_crc64(myContent)
    crc_header = conn.provider.checksum_crc64ecma_header
    k.set_contents_from_string(myContent, headers={crc_header: local_crc})


def test_multi_upload_checkingCrc64ByHeader():
    from filechunkio import FileChunkIO
    # 服务端不比对文件整体crc，可以在本地比对
    conn.enable_crc = False
    crc_header = conn.provider.checksum_crc64ecma_header
    crc_obj = utils.Crc64(0)
    final_crc = 0

    source_size = os.stat(file_path).st_size

    mp = bucket.initiate_multipart_upload("test_multi")

    chunk_size = 100 * 1024
    count = int(math.ceil(source_size * 1.0 / chunk_size * 1.0))

    for i in range(count):
        offset = chunk_size * i
        read_size = min(chunk_size, source_size - offset)
        with FileChunkIO(file_path, 'r', offset=offset, bytes=read_size) as fp:
            # 计算单个分片的crc
            local_part_crc = utils.compute_file_crc64(file_path, offset, offset + read_size - 1)
            # 将每一个分片的crc合并，最终将得到文件整体的crc值
            final_crc = crc_obj.combine(final_crc, int(local_part_crc), read_size)
            # 逐个上传分片
            mp.upload_part_from_file(fp, part_num=i + 1, headers={crc_header: local_part_crc})
    ret = mp.complete_upload()

    assert ret.ChecksumCRC64ECMA == str(final_crc)
    assert ret and ret.status == 200



from ks3.responseResult import ResponseMetadata


class BucketInventory(object):
    OPTIONAL_FIELD_SIZE = 'Size'
    OPTIONAL_FIELD_LAST_MODIFIED_DATE = 'LastModifiedDate'
    OPTIONAL_FIELD_ETAG = 'ETag'
    OPTIONAL_FIELD_STORAGE_CLASS = 'StorageClass'
    OPTIONAL_FIELD_IS_MULTIPART_UPLOADED = 'IsMultipartUploaded'
    OPTIONAL_FIELD_ENCRYPTION_STATUS = 'EncryptionStatus'

    def __init__(self, id=None, is_enabled=None, filter=None, destination=None, schedule=None,
                 optional_fields=None, *args, **kwargs):
        if optional_fields is None:
            optional_fields = []
        self.id = id
        self.is_enabled = 'true' if is_enabled else 'false'
        self.filter = filter
        self.destination = destination
        self.schedule = schedule
        self.optional_fields = optional_fields

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        if name == 'Destination':
            self.destination = Destination()
            return self.destination
        if name == 'Schedule':
            self.schedule = Schedule()
            return self.schedule
        if name == 'Filter':
            self.filter = Filter()
            return self.filter
        return None

    def endElement(self, name, value, connection):
        if name == 'Id':
            self.id = value
        if name == 'IsEnabled':
            self.is_enabled = value
        if name == 'Field':
            self.optional_fields.append(value)
        setattr(self, name, value)

    def to_xml(self):
        s = u'<InventoryConfiguration>'
        if self.id is not None:
            s += '<Id>%s</Id>' % self.id
        if self.is_enabled is not None:
            s += '<IsEnabled>%s</IsEnabled>' % self.is_enabled
        if self.destination is not None:
            s += '<Destination>%s</Destination>' % self.destination.to_xml()
        if self.schedule is not None:
            s += self.schedule.to_xml()
        if self.filter is not None:
            s += self.filter.to_xml()
        if self.optional_fields is not None:
            s += '<OptionalFields>'
            for i in self.optional_fields:
                s += '<Field>%s</Field>' % i
            s += '</OptionalFields>'
        s += '</InventoryConfiguration>'
        return s


class Destination(object):
    FORMAT_CSV = 'CSV'
    FORMAT_ORC = 'ORC'
    FORMAT_PARQUET = 'Parquet'

    def __init__(self, format=None, account_id=None, bucket=None, prefix=None):
        self.format = format
        self.account_id = account_id
        self.bucket = bucket
        self.prefix = prefix

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Format':
            self.format = value
        if name == 'AccountId':
            self.account_id = value
        if name == 'Bucket':
            self.bucket = value
        if name == 'Prefix':
            self.prefix = value
        setattr(self, name, value)

    def to_xml(self):
        s = u'<KS3BucketDestination>'
        if self.format is not None:
            s += '<Format>%s</Format>' % self.format
        if self.account_id is not None:
            s += '<AccountId>%s</AccountId>' % self.account_id
        if self.bucket is not None:
            s += '<Bucket>%s</Bucket>' % self.bucket
        if self.prefix is not None:
            s += '<Prefix>%s</Prefix>' % self.prefix
        s += '</KS3BucketDestination>'
        return s


class Filter(object):

    def __init__(self, prefix=None, last_modify_begin_time_stamp=None, last_modify_end_time_stamp=None):
        self.prefix = prefix
        self.last_modify_begin_time_stamp = last_modify_begin_time_stamp
        self.last_modify_end_time_stamp = last_modify_end_time_stamp

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Prefix':
            self.prefix = value
        if name == 'LastModifyBeginTimeStamp':
            self.last_modify_begin_time_stamp = value
        if name == 'LastModifyEndTimeStamp':
            self.last_modify_end_time_stamp = value
        setattr(self, name, value)

    def to_xml(self):
        s = u'<Filter>'
        if self.prefix is not None:
            s += '<Prefix>%s</Prefix>' % self.prefix
        if self.last_modify_begin_time_stamp is not None:
            s += '<LastModifyBeginTimeStamp>%s</LastModifyBeginTimeStamp>' % self.last_modify_begin_time_stamp
        if self.last_modify_end_time_stamp is not None:
            s += '<LastModifyEndTimeStamp>%s</LastModifyEndTimeStamp>' % self.last_modify_end_time_stamp
        s += '</Filter>'
        return s


class Schedule(object):
    ONCE = 'Once'
    WEEKLY = 'Weekly'

    def __init__(self, frequency=None):
        self.frequency = frequency

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Frequency':
            self.frequency = value
        setattr(self, name, value)

    def to_xml(self):
        s = u'<Schedule>'
        if self.frequency is not None:
            s += '<Frequency>%s</Frequency>' % self.frequency
        s += '</Schedule>'
        return s


class ListInventoryConfigurationsResult(object):
    def __init__(self, **kwargs):
        self.inventory_configurations = []

        self.response_metadata = ResponseMetadata(**kwargs)

    def startElement(self, name, attrs, connection):
        if name == 'InventoryConfiguration':
            self.inventory_configurations.append(BucketInventory())
            return self.inventory_configurations[-1]
        return None

    def endElement(self, name, value, connection):
        setattr(self, name, value)



import copy
import logging
import os
import tempfile
from concurrent import futures

from ks3.exception import KS3ClientError, S3ResponseError
from ks3.upload import sse_c_header_keys
from ks3.utils import ResumeRecordManager, compute_file_crc64, check_crc, force_rename, \
    DEFAULT_PART_SIZE
import math

logger = logging.getLogger(__name__)


class DownloadTask:
    def __init__(
            self, key, bucket, filename,
            executor,
            part_size=DEFAULT_PART_SIZE,
            resumable=False,
            resumable_filename=None,
    ):
        """
        key: 需要下载的Key对象（ks3.key.Key）
        bucket: Bucket对象（ks3.bucket.Bucket）
        filename: 下载到的本地文件路径
        executor: 线程池
        part_size: 分块大小
        resumable: 是否开启断点续传，默认False。开启断点续传时，如果本地存在有效的断点续传记录，则尝试恢复下载
        resumable_filename: 断点续传记录文件路径。如果不指定，则使用默认路径（self.filename + '.ks3resume'）
        """
        self.key = key
        self.bucket = bucket
        self.filename = filename
        self.executor = executor
        self.part_size = part_size
        self.resumable = resumable
        if self.resumable:
            if resumable_filename is None:
                resumable_filename = self.filename + '.ks3resume'
            self.record_manager = ResumeRecordManager(resumable_filename)
        else:
            self.record_manager = None
        # 下载过程中临时文件名
        self.tmp_file_name = None

    def download(self, headers=None):
        if headers is None:
            headers = {}
        # 传入的key对象没有etag属性，则认为缺少元数据信息，需要head获取
        head_headers = {}
        for k in sse_c_header_keys:
            if k in headers:
                head_headers[k] = headers[k]
        if self.key.etag is None:
            head_k, resp = self.bucket._get_key_internal(self.key.name, head_headers, None)
            if head_k is None:
                logger.error('Failed to head key {0} from bucket {1}'
                             .format(self.key.name, self.bucket.name))
                raise S3ResponseError(resp.status, resp.reason, None, request_id=resp.getheader('x-kss-request-id'))
            self.key = head_k
        self.multipart_download(headers)
        return DownloadResult(self.key.server_crc, self.key.etag)

    def multipart_download(self, headers=None):
        self._prepare_record_and_tmp_file()
        # 如果开启断点续传则获取已下载的分块信息，否则赋值为空字典
        part_downloaded = self.record_manager.record.part_infos if self.resumable else {}

        num_parts = int(math.ceil(self.key.size / float(self.part_size)))
        remain_size = self.key.size

        part_futures = {}
        for part_num in range(1, num_parts + 1):
            if part_num in part_downloaded:
                logger.debug('key_name={0}, part_num={1} already downloaded, skip'.format(self.key.name, part_num))
                remain_size -= self.part_size
                continue
            part_info = DownloadPartInfo(part_num, self.part_size * (part_num - 1), min(self.part_size, remain_size))
            part_futures[self.executor.submit(
                self._download_part,
                part_info=part_info,
                headers=headers,
            )] = part_num
            remain_size -= self.part_size
        self.executor.shutdown(wait=True)
        for future in futures.as_completed(part_futures):
            try:
                future.result()
            except Exception as e:
                logger.error(
                    'key_name={0}, part {1} download failed: {2}'.format(self.key.name, part_futures[future], e))
                raise e

        if self.bucket.connection.enable_crc:
            client_crc = compute_file_crc64(self.tmp_file_name)
            if not check_crc(client_crc, self.key.server_crc):
                if self.resumable:
                    self.record_manager.delete()
                logger.error(
                    "key_name={0}, Inconsistent CRC checksum client_crc: {1}, server_crc: {2}"
                    .format(self.key.name, client_crc, self.key.server_crc))
                raise KS3ClientError(
                    "Inconsistent CRC checksum client_crc: %s, server_crc: %s" % (client_crc, self.key.server_crc))
        force_rename(self.tmp_file_name, self.filename)
        if self.resumable:
            self.record_manager.delete()

    def _download_part(self, part_info, headers=None):
        with open(self.tmp_file_name, 'rb+') as f:
            f.seek(part_info.start, os.SEEK_SET)
            # 下载对象会读写当前key对象的resp属性，线程不安全
            # 需要使用线程独享的key对象进行下载
            byte_range = (part_info.start, part_info.start + part_info.part_size - 1)
            copy.copy(self.key).get_contents_to_file(f, byte_range=byte_range, headers=headers)
            if self.resumable:
                self.record_manager.record.part_infos[part_info.part_num] = part_info
                self.record_manager.save()

    def _prepare_record_and_tmp_file(self):
        if not self.resumable:
            self.tmp_file_name = self._create_tmp_file()
            return

        self.record_manager.load()
        if self.record_manager.record is not None and not self._check_record_valid():
            logger.debug('key_name={0}, found invalid record, delete it'.format(self.key.name))
            self.record_manager.delete()
        if self.record_manager.record is None:
            logger.debug('key_name={0}, not found record, initiate multipart download'.format(self.key.name))
            self.record_manager.record = DownloadRecord(self.key.size, self.key.last_modified, self.key.etag,
                                                        self.bucket.name, self.key.name, self.part_size,
                                                        self._create_tmp_file(), {})
            self.record_manager.save()
        self.tmp_file_name = self.record_manager.record.tmp_file_name
        self.part_size = self.record_manager.record.part_size

    def _create_tmp_file(self):
        directory_path = os.path.dirname(self.filename)
        base_name = os.path.basename(self.filename)
        with tempfile.NamedTemporaryFile(
                mode='a', dir=directory_path, prefix=base_name + '.', suffix='.ks3temp', delete=False) as tmp_fp:
            tmp_name = tmp_fp.name
        return tmp_name

    def _check_record_valid(self):
        record = self.record_manager.record
        if not isinstance(record, DownloadRecord):
            return False
        if not os.path.exists(record.tmp_file_name):
            return False
        for attr in [record.bucket_name, record.key_name, record.mtime, record.etag]:
            if not isinstance(attr, str):
                return False
        for attr in [record.file_size, record.part_size]:
            if not isinstance(attr, int):
                return False
        if not isinstance(record.part_infos, dict):
            return False
        if record.mtime != self.key.last_modified or record.file_size != self.key.size or record.etag != self.key.etag:
            return False
        return True


class DownloadRecord(object):
    def __init__(self, file_size, mtime, etag, bucket_name, key_name, part_size, tmp_file_name, part_infos):
        self.file_size = file_size
        self.mtime = mtime
        self.etag = etag
        self.bucket_name = bucket_name
        self.key_name = key_name
        self.part_size = part_size
        self.tmp_file_name = tmp_file_name
        # <part_num, DownloadPartInfo>
        self.part_infos = part_infos


class DownloadPartInfo(object):
    def __init__(self, part_num, start, part_size):
        self.part_num = part_num
        self.start = start
        self.part_size = part_size


class DownloadResult:
    def __init__(
            self,
            crc64ecma=None,
            etag=None,
    ):
        self.crc64ecma = crc64ecma
        self.etag = etag



# coding:utf-8

#  This software code is made available without warranties of any
#  kind.  You may copy, display, modify and redistribute the software
#  code either by itself or as incorporated into your code; provided that
#  you do not remove any proprietary notices. Your use of this software
#  code is at your own risk.
import os
import time
import xml.sax

import requests

from ks3 import auth
from ks3 import handler
from ks3 import utils
from ks3.auth import url_encode
from ks3.bucket import Bucket, BucketLocation
from ks3.exception import S3ResponseError, S3CreateError, KS3ClientError, ParamValidationError
from ks3.http import make_request
from ks3.http import KS3HTTPAdapter
from ks3.provider import Provider
from ks3.responseResult import ResponseResult
from ks3.resultset import ResultSet
from ks3.utils import Ks3DnsResolver, RetryPolicy

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2


dns_resolver = Ks3DnsResolver()


def check_lowercase_bucketname(n):
    """
    Bucket names must not contain uppercase characters. We check for
    this by appending a lowercase character and testing with islower().
    Note this also covers cases like numeric bucket names with dashes.

    >>> check_lowercase_bucketname("Aaaa")
    Traceback (most recent call last):
    ...
    BotoClientError: S3Error: Bucket names cannot contain upper-case
    characters when using either the sub-domain or virtual hosting calling
    format.

    >>> check_lowercase_bucketname("1234-5678-9123")
    True
    >>> check_lowercase_bucketname("abcdefg1234")
    True
    """
    if not (n + 'a').islower():
        raise KS3ClientError("Bucket names cannot contain upper-case " \
                             "characters when using either the sub-domain or virtual " \
                             "hosting calling format.")
    return True


class _CallingFormat(object):

    def get_bucket_server(self, server, bucket):
        return ''

    def build_url_base(self, connection, protocol, server, bucket, key=''):
        url_base = '%s://' % protocol
        url_base += self.build_host(server, bucket)
        url_base += connection.get_path(self.build_path_base(bucket, key))
        return url_base

    def build_host(self, server, bucket):
        if bucket == '':
            return server
        else:
            return self.get_bucket_server(server, bucket)

    def build_auth_path(self, bucket, key=''):
        key = utils.get_utf8_value(key)
        path = ''
        if bucket != '':
            path = '/' + bucket
        buf = path + '/%s' % parse.quote(key)
        buf = buf.replace('//', '/%2F')
        return buf

    def build_path_base(self, bucket, key=''):
        buf = '/%s' % url_encode(key)
        buf = buf.replace('//', '/%2F')
        return buf


class PathCallingFormat(_CallingFormat):

    def get_bucket_server(self, server, bucket):
        return server

    def build_path_base(self, bucket, key=''):
        path_base = '/'
        if bucket:
            path_base += "%s/" % bucket
        path_base += url_encode(key)
        path_base = path_base.replace('//', '/%2F')
        return path_base


class SubdomainCallingFormat(_CallingFormat):

    def get_bucket_server(self, server, bucket):
        return '%s.%s' % (bucket, server)


# 习惯上倾向于称之为VirtualHostCallingFormat，故添加此别名类
class VirtualHostCallingFormat(SubdomainCallingFormat):
    pass


class Connection(object):
    QueryString = 'Signature=%s&Expires=%d&KSSAccessKeyId=%s'

    def __init__(self, access_key_id, access_key_secret, host="",
                 port=80, provider='kss', security_token=None, profile_name=None, path='/',
                 is_secure=False, debug=0, calling_format=VirtualHostCallingFormat(), domain_mode=False,
                 local_encrypt=False, local_key_path="", timeout=10, ua_addon='', enable_crc=True, block_size=8192,
                 proxy_host=None, proxy_port=None, proxy_username=None, proxy_password=None, connection_pool_size=1024,
                 dns_cache_time=30, retry_policy=RetryPolicy(), need_auth_header=True):
        """
        :param access_key_id: 金山云提供的ACCESS KEY ID
        :param access_key_secret: 金山云提供的SECRET KEY ID
        :param host: 请参考官网API文档说明中的Region定义(https://docs.ksyun.com/read/latest/65/_book/index.html)
        :param port: 请求端口，默认80
        :param is_secure: 是否启用HTTPS，True:启用  False:关闭
        :param domain_mode: 是否使用自定义域名访问，True:是 False:否
        :param local_encrypt: 是否启用本地加密， True:是 False:否，默认False，如选是，需要配置本地密钥路径
        :param enable_crc: 是否启用crc64校验，True:是 False:否，默认True
        :param block_size: 发送请求时，从本地读数据或者向服务器发数据的数据块大小
        """
        if host is None or host.strip() == '':
            raise ParamValidationError("host cannot be blank")

        self.access_key_id = access_key_id
        self.access_key_secret = access_key_secret
        self.is_secure = is_secure
        self.host = host
        self.port = port
        self.debug = debug
        self.path = path
        self.calling_format = calling_format
        self.domain_mode = domain_mode
        self.local_encrypt = local_encrypt
        self.key = ""
        self.timeout = timeout
        self.block_size = block_size
        self.ua_addon = ua_addon
        self.enable_crc = enable_crc
        self.need_auth_header = need_auth_header
        if self.is_secure:
            self.protocol = 'https'
            if self.port == 80:
                self.port = 443
        else:
            self.protocol = 'http'

        if isinstance(provider, Provider):
            # Allow overriding Provider
            self.provider = provider
        else:
            self._provider_type = provider
            self.provider = Provider(self._provider_type,
                                     access_key_id,
                                     access_key_secret,
                                     security_token,
                                     profile_name)

        # Allow config file to override default host, port, and host header.
        if self.provider.host:
            self.host = self.provider.host
        if self.provider.port:
            self.port = self.provider.port
        if self.provider.host_header:
            self.host_header = self.provider.host_header
        if self.local_encrypt:
            self.load_key(local_key_path)
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_username = proxy_username
        self.proxy_password = proxy_password

        self.session = requests.Session()
        adapter_kwargs = {
            'pool_connections': connection_pool_size,
            'pool_maxsize': connection_pool_size,
            'max_retries': 0,
            'block_size': block_size,
        }
        self.session.mount('http://', KS3HTTPAdapter(**adapter_kwargs))
        self.session.mount('https://', KS3HTTPAdapter(**adapter_kwargs))

        if dns_cache_time > 0 and dns_cache_time is not None:
            dns_resolver.ttl = dns_cache_time
            dns_resolver.open_cache()
        self.retry_policy = retry_policy

    def load_key(self, path):
        error_msg = "In local_encrypt mode, we need you to indicate the location of your private key. Set value for 'local_key_path' while initiate connection."
        assert path, error_msg
        with open(path, 'rb') as ak_file:
            assert os.path.getsize(path), "The key file should not be empty"
            content = ak_file.read()
            assert len(content.strip()) == 16, "The key's length should be 16"
            self.key = content.strip()

    def make_request(
            self, method, bucket="", key="", data="",
            headers=None, query_args=None, metadata=None, timeout=10,
            retry_policy=None,  # 兼容list方法的重试策略
    ):
        if not headers:
            headers = {}
        if not query_args:
            query_args = {}
        if not metadata:
            metadata = {}
        timeout = self.timeout

        request_kwargs = {
            'server': self.host,
            'port': self.port,
            'access_key_id': self.access_key_id,
            'access_key_secret': self.access_key_secret,
            'bucket': bucket,
            'key': key,
            'query_args': query_args,
            'headers': headers,
            'data': data,
            'metadata': metadata,
            'method': method,
            'is_secure': self.is_secure,
            'domain_mode': self.domain_mode,
            'timeout': timeout,
            'ua_addon': self.ua_addon,
            'proxy_host': self.proxy_host,
            'proxy_port': self.proxy_port,
            'proxy_username': self.proxy_username,
            'proxy_password': self.proxy_password,
            'calling_format': self.calling_format,
            'session': self.session,
            'need_auth_header': self.need_auth_header,
        }
        if retry_policy is None:
            retry_policy = self.retry_policy
        if retry_policy is not None:
            resp = retry_policy.call(
                make_request,
                **request_kwargs
            )
        else:
            resp = make_request(**request_kwargs)

        return resp

    def get_all_buckets(self, headers=None, project_ids=None,
                        region=None, prefix=None, visit_type=None, storage_type=None):
        query_args = {}
        if project_ids is not None:
            query_args = {
                "projectIds": project_ids
            }
        response = self.make_request('GET', headers=headers, query_args=query_args)
        body = response.read()
        if response.status > 300:
            raise S3ResponseError(response.status, response.reason, body)
        result = ResultSet(marker_elem=[('Bucket', Bucket)], status=response.status, reason=response.reason,
                           headers=response.headers, raw_body=body)
        h = handler.XmlHandler(result, self)
        if not isinstance(body, bytes):
            body = body.encode('utf-8')
        xml.sax.parseString(body, h)

        # 按条件过滤
        if region is not None:
            new_result = list(filter(lambda b: b.region == region, result))
            result.clear()
            result.extend(new_result)
        if prefix is not None:
            new_result = list(filter(lambda b: b.name.startswith(prefix), result))
            result.clear()
            result.extend(new_result)
        if visit_type is not None:
            new_result = list(filter(lambda b: b.VisitType == visit_type, result))
            result.clear()
            result.extend(new_result)
        if storage_type is not None:
            new_result = list(filter(lambda b: b.type == storage_type, result))
            result.clear()
            result.extend(new_result)

        return result

    def get_bucket(self, bucket_name, headers=None):
        return Bucket(self, bucket_name)

    def head_bucket(self, bucket_name, headers=None):
        response = self.make_request('HEAD', bucket_name)
        body = response.read()
        if response.status == 200:
            return ResponseResult(response.headers, status=response.status, reason=response.reason,
                                  headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, None,
                                  request_id=response.getheader('x-kss-request-id'))

    def get_bucket_location(self, bucket_name):
        response = self.make_request('GET', bucket_name, query_args='location')
        body = response.read()
        if response.status == 200:
            loc = BucketLocation(status=response.status, reason=response.reason, headers=response.headers,
                                 raw_body=body)
            h = handler.XmlHandler(loc, self)
            xml.sax.parseString(body, h)
            return loc
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def create_bucket(self, bucket_name, headers=None, project_id=None,
                      location=None, policy=None):
        check_lowercase_bucketname(bucket_name)

        if policy:
            if headers:
                headers[self.provider.acl_header] = policy
            else:
                headers = {self.provider.acl_header: policy}
        if location is None:
            data = ''
        else:
            data = '<CreateBucketConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><LocationConstraint>' + \
                   location + '</LocationConstraint></CreateBucketConfiguration>'

        query_args = {}
        if project_id is not None:
            query_args = {
                "projectId": project_id
            }
        response = self.make_request('PUT', bucket_name, headers=headers, query_args=query_args,
                                     data=data)
        body = response.read()
        if response.status == 409:
            raise S3CreateError(response.status, response.reason, body)
        if response.status == 200:
            return Bucket(self, bucket_name, status=response.status, reason=response.reason,
                          headers=response.headers)
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def delete_bucket(self, bucket_name, headers=None):
        """
        Removes an S3 bucket.

        In order to remove the bucket, it must first be empty. If the bucket is
        not empty, an ``S3ResponseError`` will be raised.
        """
        response = self.make_request('DELETE', bucket_name, headers=headers)
        body = response.read()
        if response.status != 204:
            raise S3ResponseError(response.status, response.reason, body)
        return ResponseResult(None, status=response.status, reason=response.reason,
                              headers=response.headers)

    def generate_url(self, expires_in, method, bucket='', key='', headers=None,
                     query_auth=True, force_http=False, response_headers=None,
                     expires_in_absolute=False, version_id=None):

        headers = headers or {}
        if expires_in_absolute:
            expires = int(expires_in)
        else:
            expires = int(time.time() + expires_in)
        # auth_path = self.calling_format.build_auth_path(bucket, key)
        # auth_path = self.get_path(auth_path)
        # optional version_id and response_headers need to be added to
        # the query param list.
        extra_qp = []
        extra_qp_encoded = []
        query_args = None
        if version_id is not None:
            extra_qp.append("versionId=%s" % version_id)
        if response_headers:
            for k, v in list(response_headers.items()):
                extra_qp.append("%s=%s" % (k, v))
                extra_qp_encoded.append("%s=%s" % (k, parse.quote(v)))
        if extra_qp:
            query_args = '&'.join(extra_qp)

        # if not headers.has_key('Date'):
        #    headers['Date'] = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())

        c_string = auth.canonical_string(method, bucket, key, query_args=query_args, headers=headers, expires=expires)
        b64_hmac = auth.encode(self.access_key_secret, c_string)
        encoded_canonical = parse.quote(b64_hmac, safe='')
        if query_auth:
            encode_ak = self.access_key_id
            # encode_ak = parse.quote(self.access_key_id)
            # print 'encode_ak:%s'%encode_ak
            query_part = '?' + self.QueryString % (encoded_canonical, expires, encode_ak)
        else:
            query_part = ''
        if headers:
            hdr_prefix = self.provider.header_prefix
            for k, v in list(headers.items()):
                if k.startswith(hdr_prefix):
                    # headers used for sig generation must be
                    # included in the url also.
                    extra_qp.append("%s=%s" % (k, parse.quote(v)))
        if extra_qp_encoded:
            delimiter = '?' if not query_part else '&'
            query_part += delimiter + '&'.join(extra_qp_encoded)
        if force_http:
            protocol = 'http'
            port = 80
        else:
            protocol = self.protocol
            port = self.port
        return self.calling_format.build_url_base(self, protocol,
                                                  self.server_name(port),
                                                  bucket, key) + query_part

    def server_name(self, port=None):
        if not port:
            port = self.port
        if port == 80:
            signature_host = self.host
        else:
            signature_host = "%s:%s" % (self.host, port)
        return signature_host

    def get_path(self, path='/'):
        # The default behavior is to suppress consecutive slashes for reasons
        # discussed at
        # https://groups.google.com/forum/#!topic/boto-dev/-ft0XPUy0y8
        # You can override that behavior with the suppress_consec_slashes param.
        pos = path.find('?')
        if pos >= 0:
            params = path[pos:]
            path = path[:pos]
        else:
            params = None
        if path[-1] == '/':
            need_trailing = True
        else:
            need_trailing = False
        path_elements = self.path.split('/')
        path_elements.extend(path.split('/'))
        path_elements = [p for p in path_elements if p]
        path = '/' + '/'.join(path_elements)
        if path[-1] != '/' and need_trailing:
            path += '/'
        if params:
            path = path + params
        return path

    def get_adp(self, task_id):
        query_args = 'queryadp'
        response = self.make_request('GET', task_id, query_args=query_args)
        body = response.read()
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)
        return ResponseResult(body, status=response.status, reason=response.reason,
                              headers=response.headers)



import base64
import hmac
import logging
import time

from ks3 import utils

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2

try:
    from hashlib import sha1 as sha
    from hashlib import sha256 as sha256
except ImportError:
    import sha

    sha256 = None


logger = logging.getLogger(__name__)


# query string argument of interest
qsa_of_interest = ['acl', 'cors', 'defaultObjectAcl', 'location', 'logging',
                   'partNumber', 'policy', 'requestPayment', 'torrent',
                   'versioning', 'versionId', 'versions', 'website',
                   'uploads', 'uploadId', 'response-content-type',
                   'response-content-language', 'response-expires',
                   'response-cache-control', 'response-content-disposition',
                   'response-content-encoding', 'delete', 'lifecycle',
                   'tagging', 'restore', 'notification', 'thumbnail', 'queryadp',
                   'adp', 'asyntask', 'querytask', 'domain',
                   'storageClass',
                   'websiteConfig',
                   'compose', 'quota', 'policy', 'crr', 'fetch', 'append', 'position',
                   'mirror', 'retention', 'recycle', 'recover', 'clear', 'inventory',
                   'id', 'migration', 'bucketqos', 'requesterqos']


def url_encode(key):
    if not key:
        return ""
    encode_key = parse.quote(utils.get_utf8_value(key))
    # if '%20' in encode_key:
    #     encode_key = encode_key.replace('%20', '+')

    # if '%2A' in encode_key:
    #     encode_key = encode_key.replace('%2A', '*')

    if '%7E' in encode_key:
        encode_key = encode_key.replace('%7E', '~')

    if '%2F' in encode_key:
        encode_key = encode_key.replace('%2F', '/')

    return encode_key


def encode_params(query_args):
    if not query_args:
        return ""
    map_args = {}
    if isinstance(query_args, dict):
        map_args = query_args
    else:
        for param in query_args.split("&"):
            kv = param.split("=", 1)
            k = kv[0]
            if len(kv) == 1:
                map_args[k] = ""
            else:
                v = kv[1]
                map_args[k] = v
    if not map_args:
        return ""
    sorted_keys = list(map_args.keys())
    sorted_keys.sort()
    buf_list = []
    for k in sorted_keys:
        if k in qsa_of_interest:
            v = map_args.get(k)
            if v:
                buf_list.append("%s=%s" % (k, v))
            else:
                buf_list.append("%s" % k)
    return "&".join(buf_list)


def canonical_resource(bucket, key, query_args):
    buf = "/"
    if bucket:
        buf += "%s/" % bucket
    if key:
        buf += "%s" % url_encode(key)

    buf = buf.replace('//', '/%2F')
    params = encode_params(query_args)
    if params:
        buf += "?%s" % params
    return buf


def canonical_headers(headers):
    if not headers:
        return ""
    interesting_headers = {}
    for header_key in headers:
        lk = header_key.lower()
        if lk.startswith("x-kss-"):
            interesting_headers[lk] = headers[header_key]
    if not interesting_headers:
        return ""
    sorted_header_keys = list(interesting_headers.keys())
    sorted_header_keys.sort()
    buf_list = []
    for header_key in sorted_header_keys:
        buf_list.append("%s:%s" % (header_key, interesting_headers.get(header_key)))
    return "\n".join(buf_list)


def canonical_string(method, bucket="", key="", query_args=None, headers=None, expires=None):
    if not headers:
        headers = {}
    if not query_args:
        query_args = ""

    can_resource = canonical_resource(bucket, key, query_args)
    can_headers = canonical_headers(headers)
    content_md5 = ""
    content_type = ""
    date = ""
    for header_key in headers:
        lk = header_key.lower()
        val = headers.get(header_key)
        if not val:
            continue
        if lk == "content-md5":
            content_md5 = val
        elif lk == "content-type":
            content_type = val
        elif lk == "date":
            date = val
    if expires:
        date = str(expires)
    sign_list = [method, content_md5, content_type, date]
    if can_headers:
        sign_list.append(can_headers)
    sign_list.append(can_resource)
    sign_str = "\n".join(sign_list)
    logger.debug('sign_list: {0}'.format(sign_list))
    return sign_str


def encode(secret_access_key, str_to_encode, urlencode=False):
    secret_access_key = secret_access_key.encode('utf-8')
    str_to_encode = str_to_encode.encode('utf-8')
    b64_hmac = base64.b64encode(hmac.new(secret_access_key, str_to_encode, sha).digest()).strip().decode('utf-8')
    if urlencode:
        return parse.quote_plus(b64_hmac)
    else:
        return b64_hmac


def add_auth_header(access_key_id, secret_access_key, headers, method, bucket, key, query_args):
    if not access_key_id:
        return
    if 'Date' not in headers:
        from email.utils import formatdate
        headers['Date'] = formatdate(time.time(), usegmt=True)

    c_string = canonical_string(method, bucket, key, query_args, headers)
    headers['Authorization'] = \
        "%s %s:%s" % ("KSS", access_key_id, encode(secret_access_key, c_string))
